/*
 * Copyright 2021 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.tomee;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import org.apache.tomcat.vestige.full_installer.TomcatVestigeFullInstaller;
import org.apache.tomcat.vestige.installer.FileAdder;
import org.apache.tomcat.vestige.installer.InstallerUtils;

/**
 * @author Gael Lalire
 */
public class TomEEWebVestigeInstaller extends TomcatVestigeFullInstaller {

    private Set<String> excludes = new HashSet<String>();

    public TomEEWebVestigeInstaller(final File conf) throws IOException {
        super(conf);
        BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(TomEEWebVestigeInstaller.class.getResourceAsStream("excludes.txt"), "UTF-8"));
        String line = bufferedReader.readLine();
        while (line != null) {
            excludes.add(line.replace('/', File.separatorChar));
            line = bufferedReader.readLine();
        }
    }

    @Override
    public void getWebapps(final FileAdder fileAdder) throws IOException {
        super.getConf(new FileAdder() {

            @Override
            public void addFile(final String pathFromBase, final URL url, final boolean textFile) throws IOException {
                if (!pathFromBase.startsWith("examples")) {
                    fileAdder.addFile(pathFromBase, url, textFile);
                }
            }

            @Override
            public void addDirectory(final String pathFromBase) throws IOException {
                if (!pathFromBase.startsWith("examples")) {
                    fileAdder.addDirectory(pathFromBase);
                }
            }
        });
    }

    @Override
    public void getConf(final FileAdder fileAdder) throws IOException {
        super.getConf(new FileAdder() {

            @Override
            public void addFile(final String pathFromBase, final URL url, final boolean textFile) throws IOException {
                if (!excludes.contains(pathFromBase)) {
                    fileAdder.addFile(pathFromBase, url, textFile);
                }
            }

            @Override
            public void addDirectory(final String pathFromBase) throws IOException {
                fileAdder.addDirectory(pathFromBase);
            }
        });
        InstallerUtils.addFiles(TomEEWebVestigeInstaller.class, new FileAdder() {

            @Override
            public void addFile(final String pathFromBase, final URL url, final boolean textFile) throws IOException {
                if (pathFromBase.endsWith(".original")) {
                    fileAdder.addFile(pathFromBase, url, true);
                } else {
                    fileAdder.addFile(pathFromBase, url, textFile);
                }
            }

            @Override
            public void addDirectory(final String pathFromBase) throws IOException {
                fileAdder.addDirectory(pathFromBase);
            }
        });

        fileAdder.addFile("conf" + File.separator + "tomee.xml", TomEEWebVestigeInstaller.class.getResource("tomee.xml"), true);
    }

    public void getApps(final FileAdder fileAdder) throws IOException {
        fileAdder.addFile("apps" + File.separator + "myear.vear", TomEEWebVestigeInstaller.class.getResource("myear.vear"), true);
        fileAdder.addFile("apps" + File.separator + "vear-1.0.0.xsd", TomEEWebVestigeInstaller.class.getResource("vear-1.0.0.xsd"), true);
    }

    @Override
    public void getAll(final FileAdder fileAdder) throws IOException {
        super.getAll(fileAdder);
        getApps(fileAdder);
    }

}
