package org.apache.tomee;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Collections;
import java.util.HashMap;

import org.apache.catalina.core.StandardContext;
import org.apache.catalina.webresources.StandardRoot;
import org.apache.openejb.OpenEJBException;
import org.apache.openejb.config.DeploymentLoader;
import org.apache.openejb.jee.Application;
import org.apache.openejb.sxc.ApplicationXml;
import org.apache.tomcat.vestige.launcher.webresources.VestigeJarResourceSet;
import org.apache.xbean.finder.ResourceFinder;

import fr.gaellalire.vestige.spi.resolver.VestigeJar;
import fr.gaellalire.vestige.spi.resolver.VestigeJarEntry;

public class VestigeTomEEController {

    public void setResources(final StandardContext standardContext) {
        standardContext.setResources(new StandardRoot() {

            @Override
            public void processWebInfLib() {
                super.processWebInfLib();
                for (VestigeJar vestigeJar : webApp.vestigeWarDependencies) {
                    classResources.add(new VestigeJarResourceSet(this, "/WEB-INF/classes", vestigeJar, Collections.<VestigeJar> emptyList(), "/"));
                }
            }
        });
    }

    public AppModuleCreatorData cAM() {
        AppModuleCreatorData appModuleCreatorData;
        try {
            vestigeEar = VestigeEar.create(jarFile);

            VestigeJar vestigeEarJar = vestigeEar.getVestigeJar();

            URL baseURL = new URL("jar", "", -1, vestigeEarJar.getFile().toURI().getPath() + "!/");
            VestigeJarEntry appXML = vestigeEarJar.getEntry("META-INF/application.xml");
            if (appXML != null) {
                application = ApplicationXml.unmarshal(appXML.open());
                applicationFound = true;
            } else {
                application = new Application();
            }

            VestigeJarEntry vestigeEAREntry = vestigeEarJar.getEntry("META-INF/maven-ear.xml");
            if (vestigeEAREntry != null) {
                vestigeEARProperties.load(vestigeEAREntry.open());
            }

            appDescriptors = new HashMap<String, URL>();

            for (String descriptor : DeploymentLoader.KNOWN_DESCRIPTORS) {
                VestigeJarEntry entry = vestigeEarJar.getEntry("META-INF/" + descriptor);
                if (entry != null) {
                    String name = entry.getName();
                    URL url = new URL(baseURL, name);
                    appDescriptors.put(name, url);
                }
            }

            File file = vestigeEarJar.getFile();
            appUrl = file.toURI().toURL();
            appId = file.getAbsolutePath();

            applicationXmlUrl = null;
            finder = new ResourceFinder() {
                @Override
                public URL find(final String uri) throws IOException {
                    return vestigeEar.getURL(uri);
                }
            };
            tmpClassLoader = null;

        } catch (Exception e) {
            throw new OpenEJBException(e);
        }

    }

}
